import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;

import javax.annotation.processing.AbstractProcessor;
import javax.annotation.processing.RoundEnvironment;
import javax.annotation.processing.SupportedAnnotationTypes;
import javax.annotation.processing.SupportedSourceVersion;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.AnnotationValue;
import javax.lang.model.element.Element;
import javax.lang.model.element.ElementKind;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.TypeElement;

@SupportedAnnotationTypes("Stub")
@SupportedSourceVersion(SourceVersion.RELEASE_9)
public final class StubAnnotationProcessor extends AbstractProcessor {

	private static final FileOperations FILE_OPERATIONS = FileOperations.FILE_OPERATIONS;

	private Map<String, StubBuilder> builders = new HashMap<>();

	@Override
	public boolean process(final Set<? extends TypeElement> annotations, final RoundEnvironment roundEnv) {

		System.out.println("Processing round");

		annotations.forEach(annotation -> {
			roundEnv.getElementsAnnotatedWith(annotation).forEach(annotatedElement -> {

				if (annotatedElement.getEnclosingElement().getKind() == ElementKind.INTERFACE) {

					StubBuilder stubBuilder = null;
					final Map<String, Object> annotationElements = getAnnotationElements(annotatedElement);
					final String interfaceName = annotatedElement.getEnclosingElement().getSimpleName().toString();

					System.out.println("Processing " + interfaceName);

					if (!this.builders.containsKey(interfaceName)) {
						stubBuilder = new StubBuilder();
						this.builders.put(interfaceName, stubBuilder);
					} else {
						stubBuilder = this.builders.get(interfaceName);
					}

					stubBuilder.setClassName(interfaceName).setMethod(annotatedElement, annotationElements);
				}
			});
		});

		for (Map.Entry<String, StubBuilder> entry : this.builders.entrySet()) {
			FILE_OPERATIONS.write(this.processingEnv, entry.getValue().getInterfaceName() + "Impl",
					entry.getValue().build());
		}

		this.builders = new HashMap<>();

		return true;
	}

	private Map<String, Object> getAnnotationElements(final Element annotatedElement) {
		assert !Objects.isNull(annotatedElement);

		return annotatedElement.getAnnotationMirrors().stream().flatMap(annotationMirror -> {
			final Map<? extends ExecutableElement, ? extends AnnotationValue> elementValues = annotationMirror
					.getElementValues();

			final List<AnnoationElementHolder> results = new ArrayList<>();
			for (Map.Entry<? extends ExecutableElement, ? extends AnnotationValue> entry : elementValues.entrySet()) {

				final String key = entry.getKey().getSimpleName().toString();
				final Object value = entry.getValue().getValue();

				results.add(new AnnoationElementHolder(key, value));
			}
			return results.stream();
		}).collect(Collectors.toMap(AnnoationElementHolder::getKey, AnnoationElementHolder::getValue));
	}

	private static final class AnnoationElementHolder {

		private final String key;
		private final Object value;

		AnnoationElementHolder(final String key, final Object value) {
			this.key = key;
			this.value = value;
		}

		String getKey() {
			return this.key;
		}

		Object getValue() {
			return this.value;
		}
	}
}
